from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from textwrap import dedent


class RMSProjectNotFoundError(OSError):
    """Raised when attempting to open an RMS project that does not exist."""


@dataclass
class RMSMaster:
    raw_version: str
    version: str
    fileversion: str
    variant: str
    user: str
    date: str
    time: str


def _sanitize_version(version: str) -> str:
    """Get complete RMS version.

    For RMS 10.0.0 and 11.0.0 (etc) the release is reported as
    10 or 11 in the .master file. Extend this to always have 3
    fields e.g. 10 --> 10.0.0.

    For version 14, there is now a V in front
    """

    if version[0].upper() == "V":  # e.g. version V14.1
        version = version[1:]

    # Valid for beta versions:
    if not version[-1].isdigit():
        return version

    numdots = version.count(".")
    if numdots == 0:
        return f"{version}.0.0"
    if numdots == 1:
        return f"{version}.0"
    return version


def _parse_master_file_header(filepath: Path) -> RMSMaster:
    """The root .master file creates an entry whenever there is a save event. This
    function does a light parsing of each save log entry, which are in the following
    format:

        Begin GEOMATIC file header
        ...
        date(1395)                              = 2022.09.08
        time(1395)                              = 10:58:55
        user(1395)                              = jriv
        release(1395)                           = 13.0.3
        operation(1395)                         = Save
        description(1395)                       =
        branch(1395)                            = 13_0
        build(1395)                             = 833
        variant(1395)                           = linux-amd64-gcc_4_8-release
        ...
        elements                                = 29
        filetype                                = BINARY
        fileversion                             = 2021.0000
        End GEOMATIC file header

    The number in parantheses indicates the n'th save event. The "release" is the RMS
    version that users are familiar with, hence it is renamed to version with special
    handling. 'fileversion' indicates the version of the .master file formats.
    """

    data = {
        "date": "unknown",
        "time": "unknown",
        "user": "unknown",
        "raw_version": "unknown",
        "version": "uknown",
        "variant": "unknown",
        "fileversion": "unknown",
    }

    with open(filepath, encoding="utf-8") as f:
        for line in f:
            # We only need to read the header
            if line.startswith("End GEOMATIC"):
                break

            split_line = line.split()
            if line.startswith("release"):
                data["raw_version"] = split_line[2]
                data["version"] = _sanitize_version(split_line[2])

            for key in data:
                if line.startswith(key) and len(split_line) == 3:
                    data[key] = split_line[2]

    return RMSMaster(**data)


@dataclass
class RMSProject:
    path: Path
    locked: bool
    lockfile: str | None
    master: RMSMaster

    @property
    def name(self) -> str:
        return self.path.name

    @classmethod
    def from_filepath(cls, project: str) -> RMSProject:
        project_path = Path(project)
        if not project_path.is_dir():
            raise RMSProjectNotFoundError(
                f"The project: {project_path} does not exist as a directory."
            )

        master_file = project_path / ".master"
        if not master_file.exists():
            raise FileNotFoundError(
                dedent(f"""
                RMS project .master file not found at {master_file}.
                Possible causes:
                 * The project does not exist
                 * The project data has been corrupted
                 * The project is of an unsupported version, i.e. RMS 2013.0.0x
                """)
            )
        master = _parse_master_file_header(master_file)

        lockfile = project_path / "project_lock_file"
        lockfile_content: str | None = None
        locked = lockfile.exists()
        if locked:
            with open(lockfile) as f:
                lockfile_content = f.read().strip()

        return cls(
            path=project_path,
            locked=locked,
            lockfile=lockfile_content,
            master=master,
        )



from __future__ import annotations

import json
import os
import stat
import subprocess
from unittest.mock import Mock

import pytest

from runrms.config import FMRMSConfig
from runrms.executor import FMRMSExecutor, RMSRuntimeError


def _create_config(  # noqa: PLR0913 Too many arguments in function definition (8 > 5)
    iens: int,
    run_path: str,
    project: str,
    workflow: str,
    allow_no_env: bool,
    config_file: str,
    version: str = "14.2.2",
    target_file: str | None = None,
) -> FMRMSConfig:
    args = Mock()
    args.iens = iens
    args.run_path = run_path
    args.project = project
    args.workflows = [workflow]
    args.version = version
    args.readonly = False
    args.import_path = "import/path"
    args.export_path = "export/path"
    args.allow_no_env = allow_no_env
    args.target_file = target_file
    args.setup = config_file
    args.threads = 1

    config = FMRMSConfig(args)
    config._site_config.exe = f"{os.getcwd()}/bin/rms"
    return config


def test_run_class(fm_executor_env, capsys):
    action = {"exit_status": 0}
    with open("run_path/action.json", "w") as f:
        f.write(json.dumps(action))

    config = _create_config(
        iens=0,
        project="project",
        workflow="workflow",
        run_path="run_path",
        allow_no_env=True,
        config_file="runrms.yml",
    )
    rms = FMRMSExecutor(config)
    rms.run()

    # -----------------------------------------------------------------

    action = {"exit_status": 1}
    with open("run_path/action.json", "w") as f:
        f.write(json.dumps(action))

    workflow_log = rms.config.run_path / "workflow.log"
    workflow_log.touch()
    rms_log = rms.config.run_path / "2024_RMS.log"
    rms_log.touch()

    assert rms.run() == 1
    captured = capsys.readouterr()
    assert "failed with exit status: 1. Typically this means" in captured.err
    assert f"* {workflow_log.resolve()}\n* {rms_log.resolve()}" in captured.err

    # -----------------------------------------------------------------

    action = {"exit_status": 0}
    with open("run_path/action.json", "w") as f:
        f.write(json.dumps(action))

    config = _create_config(
        iens=0,
        project="project",
        workflow="workflow",
        run_path="run_path",
        allow_no_env=True,
        target_file="some_file",
        config_file="runrms.yml",
    )

    rms = FMRMSExecutor(config)
    with pytest.raises(RMSRuntimeError) as e:
        rms.run()
        assert e.match("target-file")

    # -----------------------------------------------------------------

    action = {
        "exit_status": 0,
        "target_file": os.path.join(fm_executor_env, "some_file"),
    }
    with open("run_path/action.json", "w") as f:
        f.write(json.dumps(action))

    config = _create_config(
        iens=0,
        project="project",
        workflow="workflow",
        run_path="run_path",
        allow_no_env=True,
        target_file="some_file",
        config_file="runrms.yml",
    )

    rms = FMRMSExecutor(config)
    rms.run()


@pytest.mark.parametrize(
    "val, carry_over",
    [
        ("    ", False),
        ("", False),
        (None, False),
        ("SOME_VAL", True),
    ],
)
def test_rms_load_nonempty_exec_env_values(val, carry_over, fm_executor_env):
    # This function is to be removed when exec_env support is dropped.
    rms_exec = "runrms"
    with open(f"{rms_exec}_exec_env.json", "w") as f:
        json.dump({"RMS_TEST_VAR": val}, f)

    action = {"exit_status": 0}
    with open("run_path/action.json", "w") as f:
        f.write(json.dumps(action))

    subprocess.check_call(
        [
            rms_exec,
            "project",
            "--batch",
            "workflow",
            "--run-path",
            "run_path",
            "--iens",
            "0",
            "--version",
            "14.2.2",
            "--import-path",
            "./",
            "--export-path",
            "./",
            "--allow-no-env",
            "--setup",
            "runrms.yml",
        ]
    )

    with open("run_path/env.json") as f:
        env = json.load(f)

    if carry_over:
        assert "RMS_TEST_VAR" in env
    else:
        assert "RMS_TEST_VAR" not in env


def test_run_class_with_existing_target_file(fm_executor_env):
    target_file = os.path.join(fm_executor_env, "rms_target_file")
    action = {
        "exit_status": 0,
        "target_file": target_file,
    }
    with open("run_path/action.json", "w") as f:
        f.write(json.dumps(action))

    with open(target_file, "w") as f:
        f.write("This is a dummy target file")

    config = _create_config(
        iens=0,
        project="project",
        workflow="workflow",
        run_path="run_path",
        allow_no_env=True,
        target_file=target_file,
        config_file="runrms.yml",
    )

    rms = FMRMSExecutor(config)
    rms.run()


def test_run_wrapper(fm_executor_env, monkeypatch, capsys):
    wrapper_file_name = f"{fm_executor_env}/bin/rms_wrapper"
    with open(wrapper_file_name, "w") as f:
        f.write("#!/bin/bash\n")
        f.write("exec ${@:1}\n")
    st = os.stat(wrapper_file_name)
    os.chmod(wrapper_file_name, st.st_mode | stat.S_IEXEC)

    monkeypatch.setenv("PATH", f"{fm_executor_env}/bin:{os.environ['PATH']}")

    action = {"exit_status": 0}
    with open("run_path/action.json", "w") as f:
        f.write(json.dumps(action))

    config = _create_config(
        iens=0,
        project="project",
        workflow="workflow",
        run_path="run_path",
        allow_no_env=True,
        config_file="runrms.yml",
    )

    rms = FMRMSExecutor(config)
    rms.run()
    assert rms.run() == 0

    # -----------------------------------------------------------------

    action = {"exit_status": 1}
    with open("run_path/action.json", "w") as f:
        f.write(json.dumps(action))

    (rms.config.run_path / "workflow.log").touch()

    assert rms.run() == 1
    captured = capsys.readouterr()
    assert "failed with exit status: 1. Typically this means" in captured.err

    # -----------------------------------------------------------------

    action = {"exit_status": 0}
    with open("run_path/action.json", "w") as f:
        f.write(json.dumps(action))

    config = _create_config(
        iens=0,
        project="project",
        workflow="workflow",
        run_path="run_path",
        allow_no_env=True,
        target_file="some_file",
        config_file="runrms.yml",
    )

    rms = FMRMSExecutor(config)
    with pytest.raises(RMSRuntimeError):
        rms.run()

    # -----------------------------------------------------------------

    action = {
        "exit_status": 0,
        "target_file": os.path.join(fm_executor_env, "some_file"),
    }
    with open("run_path/action.json", "w") as f:
        f.write(json.dumps(action))

    rms.run()


def test_run_version_env(test_env_wrapper, fm_executor_env, monkeypatch):
    wrapper_file_name = f"{fm_executor_env}/bin/rms_wrapper"
    with open(wrapper_file_name, "w") as f:
        rms_wrapper = test_env_wrapper(
            expected_path_prefix="/some/path",
            expected_pythonpath="/abc/pythonpath",
        )
        f.write(rms_wrapper)

    st = os.stat(wrapper_file_name)
    os.chmod(wrapper_file_name, st.st_mode | stat.S_IEXEC)
    monkeypatch.setenv("PATH", f"{fm_executor_env}/bin:{os.environ['PATH']}")

    action = {
        "exit_status": 0,
        "target_file": os.path.join(fm_executor_env, "some_file"),
    }
    with open("run_path/action.json", "w") as f:
        f.write(json.dumps(action))

    config = _create_config(
        iens=0,
        project="project",
        workflow="workflow",
        run_path="run_path",
        allow_no_env=False,
        version="14.2.2",
        target_file="some_file",
        config_file="runrms.yml",
    )

    rms = FMRMSExecutor(config)
    rms.run()


def test_pythonpath_carried_over_from_pre_env(
    test_env_wrapper, fm_executor_env, mocker, monkeypatch
) -> None:
    monkeypatch.setenv("PYTHONPATH", "/abc/def")
    with open(f"{fm_executor_env}/bin/disable_foo", "w", encoding="utf-8") as f:
        disable_foo = test_env_wrapper(
            expected_path_prefix="/baz/bin:/foo/bin",
            expected_pythonpath="/abc/def",
            expected_lm_license_file="foo.lic",
        )
        f.write(disable_foo)
    with open("rms_exec_env.json", "w") as f:
        f.write(
            """\
{
    "PATH_PREFIX" : "/baz/bin",
    "PYTHONPATH" : "/baz/site-packages"
}
"""
        )
    mocker.patch("sys.argv", ["bin/rms"])

    action = {
        "exit_status": 0,
        "target_file": os.path.join(fm_executor_env, "some_file"),
    }
    with open("run_path/action.json", "w") as f:
        f.write(json.dumps(action))

    config = _create_config(
        iens=0,
        project="project",
        workflow="workflow",
        run_path="run_path",
        allow_no_env=False,
        version="14.2.2",
        target_file="some_file",
        config_file="runrms.yml",
    )

    rms = FMRMSExecutor(config)
    rms.run()

    with open("run_path/env.json") as f:
        env = json.load(f)

    assert env["PYTHONPATH"] == "/baz/site-packages:/foo/bar/site-packages:/abc/def"


def test_run_version_env_with_user_env(
    test_env_wrapper, fm_executor_env, mocker
) -> None:
    """Tests that a user execution environment specific in rms_exec_env.json is
    prepended to the environment already given from the configuration."""
    with open(f"{fm_executor_env}/bin/disable_foo", "w", encoding="utf-8") as f:
        disable_foo = test_env_wrapper(
            expected_path_prefix="/baz/bin:/foo/bin",
        )
        f.write(disable_foo)
    with open("rms_exec_env.json", "w") as f:
        f.write(
            """\
{
    "PATH_PREFIX" : "/baz/bin",
    "PYTHONPATH" : "/baz/site-packages"
}
"""
        )
    mocker.patch("sys.argv", ["bin/rms"])

    action = {
        "exit_status": 0,
        "target_file": os.path.join(fm_executor_env, "some_file"),
    }
    with open("run_path/action.json", "w") as f:
        f.write(json.dumps(action))

    config = _create_config(
        iens=0,
        project="project",
        workflow="workflow",
        run_path="run_path",
        allow_no_env=False,
        version="14.2.2",
        target_file="some_file",
        config_file="runrms.yml",
    )

    rms = FMRMSExecutor(config)
    rms.run()


def test_user_rms_plugins_library_env_var_preferred(
    test_env_wrapper, fm_executor_env, mocker
) -> None:
    """Tests that if a user sets the RMS_PLUGINS_LIBRARY environment variable in their
    ERT configuration, it is preferred alone over the one in the site configuration."""
    with open(f"{fm_executor_env}/bin/disable_foo", "w", encoding="utf-8") as f:
        disable_foo = test_env_wrapper(
            expected_path_prefix="/baz/bin:/foo/bin",
        )
        f.write(disable_foo)
    with open("rms_exec_env.json", "w") as f:
        f.write(
            """\
{
    "PATH_PREFIX": "/baz/bin",
    "PYTHONPATH": "/baz/site-packages",
    "RMS_PLUGINS_LIBRARY": "/user/foo"
}
"""
        )
    mocker.patch("sys.argv", ["bin/rms"])

    action = {
        "exit_status": 0,
        "target_file": os.path.join(fm_executor_env, "some_file"),
    }
    with open("run_path/action.json", "w") as f:
        f.write(json.dumps(action))

    config = _create_config(
        iens=0,
        project="project",
        workflow="workflow",
        run_path="run_path",
        allow_no_env=False,
        version="14.2.2",
        target_file="some_file",
        config_file="runrms.yml",
    )

    rms = FMRMSExecutor(config)
    rms.run()


def test_run_allow_no_env(fm_executor_env, monkeypatch):
    monkeypatch.setenv("PATH", f"{fm_executor_env}/bin:{os.environ['PATH']}")

    action = {
        "exit_status": 0,
        "target_file": os.path.join(fm_executor_env, "some_file"),
    }
    with open("run_path/action.json", "w") as f:
        f.write(json.dumps(action))

    config = _create_config(
        iens=0,
        project="project",
        workflow="workflow",
        run_path="run_path",
        allow_no_env=False,
        version="14.2.2",
        target_file="some_file",
        config_file="runrms.yml",
    )

    config._version_given = "non-existing"
    rms = FMRMSExecutor(config)
    with pytest.raises(RMSRuntimeError, match="non-existing"):
        rms.run()

    config = _create_config(
        iens=0,
        project="project",
        workflow="workflow",
        run_path="run_path",
        allow_no_env=True,
        version="14.2.2",
        target_file="some_file",
        config_file="runrms.yml",
    )

    rms = FMRMSExecutor(config)
    rms.run()


def test_rms_job_script_parser(fm_executor_env, monkeypatch):
    monkeypatch.setenv("RMS_TEST_VAR", "fdsgfdgfdsgfds")

    action = {"exit_status": 0}
    with open("run_path/action.json", "w") as f:
        f.write(json.dumps(action))

    rms_exec = "runrms"
    subprocess.check_call(
        [
            rms_exec,
            "project",
            "--batch",
            "workflow",
            "--run-path",
            "run_path",
            "--iens",
            "0",
            "--version",
            "14.2.2",
            "--import-path",
            "./",
            "--export-path",
            "./",
            "--setup",
            "runrms.yml",
        ]
    )

    subprocess.check_call(
        [
            rms_exec,
            "project",
            "-batch",
            "workflow",
            "--run-path",
            "run_path",
            "--iens",
            "0",
            "--version",
            "14.2.2",
            "--allow-no-env",
            "--setup",
            "runrms.yml",
        ]
    )


@pytest.mark.parametrize("exit_status", [1, 2, 137])
def test_print_failure_when_no_logs_found_in_rms_model(
    exit_status, fm_executor_env, capsys
):
    action = {"exit_status": exit_status}
    with open("run_path/action.json", "w") as f:
        f.write(json.dumps(action))

    config = _create_config(
        iens=0,
        project="project",
        workflow="workflow",
        run_path="run_path",
        allow_no_env=True,
        config_file="runrms.yml",
    )
    rms = FMRMSExecutor(config)
    rms.run()

    assert rms.run() == exit_status
    captured = capsys.readouterr()
    assert f"failed with exit status: {exit_status}" in captured.err
    if exit_status == 137:
        assert (
            "This often means that the compute node ran out of memory" in captured.err
        )
    else:
        assert f"* {fm_executor_env}/run_path" in captured.err
        assert "This may mean that the compute node ran out of memory" in captured.err

    for line in captured.err.split("\n"):
        assert line.startswith("\t") is False


def test_print_failure_when_logs_found_in_rms_model(fm_executor_env, capsys):
    action = {"exit_status": 1}
    with open("run_path/action.json", "w") as f:
        f.write(json.dumps(action))

    config = _create_config(
        iens=0,
        project="project",
        workflow="workflow",
        run_path="run_path",
        allow_no_env=True,
        config_file="runrms.yml",
    )
    (fm_executor_env / "run_path" / "2025_RMS.log").touch()
    (fm_executor_env / "run_path" / "workflow.log").touch()

    rms = FMRMSExecutor(config)
    rms.run()

    assert rms.run() == 1
    captured = capsys.readouterr()
    print(captured.err)
    assert "failed with exit status: 1" in captured.err
    assert f"* {fm_executor_env}/run_path" in captured.err
    assert "workflow.log" in captured.err
    assert "2025_RMS.log" in captured.err

    for line in captured.err.split("\n"):
        assert line.startswith("\t") is False


def test_lm_license_server_overwritten_during_batch(fm_executor_env) -> None:
    action = {"exit_status": 0}
    with open("run_path/action.json", "w") as f:
        f.write(json.dumps(action))

    config = _create_config(
        iens=0,
        project="project",
        workflow="workflow",
        run_path="run_path",
        allow_no_env=True,
        version="14.2.2",
        config_file="runrms.yml",
    )

    rms = FMRMSExecutor(config)
    rms.run()

    with open("run_path/env.json") as f:
        env = json.load(f)

    assert env["LM_LICENSE_FILE"] == "/license/file.lic"



from __future__ import annotations

import os
import pathlib
import shutil
import stat
from pathlib import Path
from textwrap import dedent
from typing import TYPE_CHECKING, Any

import pytest
import yaml

from runrms.config import DEFAULT_CONFIG_FILE

if TYPE_CHECKING:
    from collections.abc import Callable


@pytest.fixture(scope="session")
def source_root(request) -> Path:
    return request.config.rootpath


@pytest.fixture
def base_ert_rms_config() -> str:
    return dedent(
        r"""
        DEFINE <USER>               user
        DEFINE <SCRATCH>            $RUNRMS_TMP_PATH/scratch
        DEFINE <CASE_DIR>           snakeoil

        DEFINE <RMS_VERSION>        14.2.2
        DEFINE <RMS_NAME>           snakeoil.rms14.2.2

        NUM_REALIZATIONS 1

        QUEUE_SYSTEM LOCAL

        RANDOM_SEED 123456

        RUNPATH  <SCRATCH>/<USER>/<CASE_DIR>/realization-<IENS>/iter-<ITER>/

        FORWARD_MODEL RMS(<IENS>=<IENS>, <RMS_VERSION>=<RMS_VERSION>, <RMS_PROJECT>=<CONFIG_PATH>/../../rms/model/<RMS_NAME>)
    """  # noqa: E501
    )


@pytest.fixture
def fmu_snakeoil_project(tmp_path, monkeypatch, base_ert_rms_config) -> None:
    """Makes a skeleton FMU project structure into a tmp_path, with a basic ERT config
    that can be appended onto."""
    monkeypatch.setenv("RUNRMS_TMP_PATH", str(tmp_path))

    os.makedirs(tmp_path / "eclipse/model")
    for app in ("ert", "rms"):
        os.makedirs(tmp_path / f"{app}/bin")
        os.makedirs(tmp_path / f"{app}/input")
        os.makedirs(tmp_path / f"{app}/model")
    os.makedirs(tmp_path / "ert/input/distributions")

    pathlib.Path(tmp_path / "ert/model/snakeoil.ert").write_text(
        base_ert_rms_config, encoding="utf-8"
    )


@pytest.fixture
def create_multi_seed_file(tmp_path) -> Callable[[str], None]:
    """Returns a function for creating a multi seed file with the given content. The
    ert/input/distributions dir needs to already exist"""

    def _create_multi_seed_file(contents: str) -> None:
        pathlib.Path(tmp_path / "ert/input/distributions/random.seeds").write_text(
            contents
        )

    return _create_multi_seed_file


@pytest.fixture
def test_env_wrapper() -> Callable[[str, str, str], str]:
    def _test_env_wrapper(
        expected_path_prefix: str = "/foo/bin",
        expected_pythonpath: str = "",
        expected_rms_plugins: str = "",
        expected_lm_license_file: str = "",
    ) -> str:
        return dedent(f"""
            #!/bin/bash
            PATH_PREFIX_EXPECTED={expected_path_prefix}
            if [[ $PATH_PREFIX != $PATH_PREFIX_EXPECTED ]]
            then
                echo "PATH_PREFIX set incorrectly"
                echo $PATH_PREFIX should be $PATH_PREFIX_EXPECTED
                exit 1
            fi
            PYPATH_EXPECTED={expected_pythonpath}
            if [[ $PYTHONPATH != $PYPATH_EXPECTED ]]
            then
                echo "PYTHONPATH set incorrectly"
                echo $PYTHONPATH should be $PYPATH_EXPECTED
                exit 1
            fi
            RMS_PLUGINS_LIBRARY_EXPECTED={expected_rms_plugins}
            if [[ $RMS_PLUGINS_LIBRARY != $RMS_PLUGINS_LIBRARY_EXPECTED ]]
            then
                echo "RMS_PLUGINS_LIBRARY set incorrectly"
                echo $RMS_PLUGINS_LIBRARY should be $RMS_PLUGINS_LIBRARY_EXPECTED
                exit 1
            fi
            LM_LICENSE_FILE_EXPECTED={expected_lm_license_file}
            if [[ $LM_LICENSE_FILE != $LM_LICENSE_FILE_EXPECTED ]]
            then
                echo "LM_LICENSE_FILE set incorrectly"
                echo $LM_LICENSE_FILE should be $LM_LICENSE_FILE_EXPECTED
                exit 1
            fi
            $@
        """).strip()

    return _test_env_wrapper


@pytest.fixture
def _env_setup(tmp_path, source_root, simple_runrms_yml, monkeypatch) -> Path:
    os.chdir(tmp_path)
    os.mkdir("run_path")
    os.mkdir("bin")
    os.mkdir("project")

    path_master = Path("project/.master")
    Path(path_master).touch()
    with open(path_master, "w", encoding="utf-8") as f:
        f.write(master_version())

    shutil.copy(source_root / "tests/bin/rms", "bin")
    exe_path = f"{os.getcwd()}/bin"
    with open("runrms.yml", "w", encoding="utf-8") as f:
        f.write(simple_runrms_yml(exe_path))

    path = os.getenv("PATH", "")
    monkeypatch.setenv("PATH", os.pathsep.join([exe_path, path]))

    return tmp_path


@pytest.fixture
def executor_env(_env_setup, test_env_wrapper) -> Path:
    disable_foo = Path("bin/disable_foo")
    with open(disable_foo, "w", encoding="utf-8") as f:
        f.write(test_env_wrapper())
    disable_foo.chmod(disable_foo.stat().st_mode | stat.S_IEXEC)

    return _env_setup


@pytest.fixture
def fm_executor_env(_env_setup, test_env_wrapper, monkeypatch) -> Path:
    """This sets LM_LICENSE_FILE to mock the way a wrapper might do so, either in the
    wrapper or the executable (which can be a wrapper). The forward model can set this
    environment variable to a different file or server, which can accomplish load
    balancing."""
    license_file_from_wrapper = "foo.lic"
    monkeypatch.setenv("LM_LICENSE_FILE", license_file_from_wrapper)
    disable_foo = Path("bin/disable_foo")
    with open(disable_foo, "w", encoding="utf-8") as f:
        f.write(test_env_wrapper(expected_lm_license_file=license_file_from_wrapper))
    disable_foo.chmod(disable_foo.stat().st_mode | stat.S_IEXEC)

    return _env_setup


@pytest.fixture
def default_config_file() -> dict[str, Any]:
    with open(DEFAULT_CONFIG_FILE, encoding="utf-8") as f:
        return yaml.safe_load(f)


@pytest.fixture
def simple_runrms_yml() -> Callable[[str | Path], str]:
    def _simple_runrms_yml(exe_path: str | Path) -> str:
        return dedent(f"""
            wrapper: disable_foo
            default: 14.2.2
            exe: {exe_path}/rms
            batch_lm_license_file: /license/file.lic
            env:
              PATH_PREFIX: /foo/bin
              RMS_IPL_ARGS_TO_PYTHON: 1

            versions:
              14.2.2:
                env:
                  PYTHONPATH: /foo/bar/site-packages
                  RMS_PLUGINS_LIBRARY: /foo/plugins
                  TCL_LIBRARY: /foo/tcl
                  TK_LIBRARY: /foo/tcl
          """)

    return _simple_runrms_yml


@pytest.fixture
def simple_runrms_config(simple_runrms_yml) -> dict[str, str]:
    return yaml.safe_load(simple_runrms_yml("."))


@pytest.fixture
def patch_argv(mocker) -> Callable[[list[str]], None]:
    def _patch_argv(options: list[str]) -> None:
        mocker.patch("sys.argv", ["runrms"] + options)

    return _patch_argv


def master_version():
    return dedent(
        """
                Begin GEOMATIC file header
                release(1395)                           = 14.2.2
                End GEOMATIC file header
                """
    )



from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING

from ._rms_config import RMSConfig

if TYPE_CHECKING:
    import argparse

logger = logging.getLogger(__name__)


class InteractiveRMSConfig(RMSConfig):
    """A class which holds the necessary configuration for executing
    runrms in interactive mode.

    It is not likely that several instances of the class is required; the
    use of a class here is more for the convinience that 'self' can hold the
    different variables (attributes) across the methods.
    """

    def __init__(self, args: argparse.Namespace) -> None:
        project = os.path.abspath(args.project) if args.project else None
        if project is None and args.theproject:
            project = os.path.abspath(args.theproject)
        super().__init__(config_path=args.setup, version=args.version, project=project)

        self._threads = args.threads
        self._readonly = args.readonly
        self._workflow = args.workflow
        self._debug = args.debug
        self._dryrun = args.dryrun
        self._dpi_scaling = args.sdpi

        for key, value in vars(args).items():
            logger.debug("Arg = %s: %s", key, value)

    @property
    def debug(self) -> bool:
        return self._debug

    @property
    def dryrun(self) -> int:
        return self._dryrun

    @property
    def threads(self) -> int:
        return self._threads

    @property
    def readonly(self) -> bool:
        return self._readonly

    @property
    def workflow(self) -> str | None:
        return self._workflow



"""Test runrms script, but manual interactive testing is also needed."""

from __future__ import annotations

import datetime
import getpass
import os
import socket
import stat
import subprocess
from pathlib import Path

import pytest
import yaml

from runrms.__main__ import get_parser
from runrms.config import DEFAULT_CONFIG_FILE, InteractiveRMSConfig
from runrms.executor import InteractiveRMSExecutor

TESTRMS1 = "tests/testdata/rms/drogon.rms12.0.2"
TESTRMS2 = "tests/testdata/rms/drogon.rms13.0.3"


def test_config_init_no_project():
    args = get_parser().parse_args(["--dryrun", "--setup", DEFAULT_CONFIG_FILE])
    config = InteractiveRMSConfig(args)
    assert config.project is None
    assert config.dryrun is True
    assert config.site_config_file == DEFAULT_CONFIG_FILE


@pytest.mark.parametrize("project", [TESTRMS1, TESTRMS2])
def test_config_init_projects(source_root, project):
    project_str = str(source_root / project)
    args = get_parser().parse_args(
        [project_str, "--dryrun", "--setup", DEFAULT_CONFIG_FILE]
    )
    config = InteractiveRMSConfig(args)
    assert config.project.path == source_root / project
    assert config.dryrun is True
    assert config.site_config_file == DEFAULT_CONFIG_FILE


@pytest.mark.integration
def test_integration():
    """Test that the endpoint is installed."""
    assert subprocess.check_output(["runrms", "-h"])


def test_rms_version_from_project(source_root, tmp_path):
    """Scan master files in RMS."""
    os.chdir(tmp_path)
    args = get_parser().parse_args([str(source_root / TESTRMS1)])
    config = InteractiveRMSConfig(args)
    assert config.project.master.version == "12.0.2"


def test_runlogger_writes_to_configured_usage_log(source_root, tmp_path):
    """Tests that the 'interactive_usage_log' site configuration options works."""
    os.chdir(tmp_path)
    runrms_usage = Path(tmp_path / "runrms_usage.log").resolve()
    runrms_usage.touch()

    with open(DEFAULT_CONFIG_FILE, encoding="utf-8") as f:
        config = yaml.safe_load(f)

    config["interactive_usage_log"] = str(runrms_usage)
    # Just allow these to be resolved, not relevant to test.
    config["wrapper"] = "/bin/echo"
    config["exe"] = "/bin/echo"

    with open(tmp_path / "runrms.yml", "w", encoding="utf-8") as f:
        yaml.safe_dump(config, f)

    args = get_parser().parse_args(
        [str(source_root / TESTRMS1), "--setup", f"{tmp_path}/runrms.yml"]
    )
    config = InteractiveRMSConfig(args)
    executor = InteractiveRMSExecutor(config)
    executor._exec_rms()
    executor.runlogger()
    with open(runrms_usage, encoding="utf-8") as f:
        log_lines = f.readlines()
    assert len(log_lines) == 1

    log = log_lines[0].rstrip().split(",")
    assert log[0] == datetime.datetime.now().strftime("%Y-%m-%d")
    # Skip wall time
    assert log[2] == getpass.getuser()
    assert log[3] == socket.gethostname()
    assert log[4] == "client"
    assert log[5] == "/bin/echo"
    assert log[6] == f"/bin/echo -v 12.0.2 -project {source_root / TESTRMS1}"

    # Ensure it appends
    executor.runlogger()
    with open(runrms_usage, encoding="utf-8") as f:
        log_lines = f.readlines()
    assert len(log_lines) == 2


@pytest.mark.xfail(reason="The executable disable_komodo_exec is not available")
def test_runrms_disable_komodo_exec(tmp_path, monkeypatch):
    """Testing integration with Komodo."""
    os.chdir(tmp_path)
    Path("rms_fake").write_text(
        """\
#!/usr/bin/env python3
import os
import sys

errors = []

BACKUPS_check=set([
    "PATH",
    "KOMODO_RELEASE",
    "MANPATH",
    "LD_LIBRARY_PATH",
    "PYTHONPATH"
])
BACKUPS = set(os.environ["BACKUPS"].split(":"))

if BACKUPS != BACKUPS_check:
    errors.append(f"BACKUP error: {BACKUPS} not equal to {BACKUPS_check}")

for backup in BACKUPS:
    if f"{backup}_BACKUP" not in os.environ:
        errors.append(f"The backup for {backup} is not set")

PATH = os.environ["PATH"]
PATH_PREFIX = os.environ["PATH_PREFIX"]
if PATH.split(":")[0] != PATH_PREFIX:
    errors.append(f"PATH_PREFIX ({PATH_PREFIX}), was not prepended to PATH ({PATH})")
if PATH_PREFIX != "/some/bin/path":
    errors.append(f"The path for run_external is not corrent {PATH_PREFIX}")

if "KOMODO_RELEASE" in os.environ:
    errors.append(f"komodo release set: {os.environ['KOMODO_RELEASE']}")

if errors:
    for e in errors:
        print(e)
    sys.exit(1)
sys.exit(0)
"""
    )

    st = os.stat("rms_fake")
    os.chmod("rms_fake", st.st_mode | stat.S_IEXEC)
    monkeypatch.setenv("KOMODO_RELEASE", f"{os.getcwd()}/bleeding")
    monkeypatch.setenv("_PRE_KOMODO_MANPATH", "some/man/path")
    monkeypatch.setenv("_PRE_KOMODO_LD_LIBRARY_PATH", "some/ld/path")

    args = get_parser().parse_args(["-v", "13.0.3"])
    config = InteractiveRMSConfig(args)

    config.path_prefix = "/some/bin/path"
    config.version_requested = "13.0.3"
    config.exe = "./rms_fake"
    config.pythonpath = ""
    config.pluginspath = "rms/plugins/path"

    executor = InteractiveRMSExecutor(config)

    return_code = executor.run()
    assert return_code == 0



from __future__ import annotations

from unittest.mock import Mock

import pytest

from runrms.config import (
    DEFAULT_CONFIG_FILE,
    FMRMSConfig,
    RMSProjectNotFoundError,
)


def _mocked_args() -> Mock:
    args = Mock()
    args.iens = -1
    args.project = "project"
    args.workflow = "workflow"
    args.run_path = "run_path"
    args.version = "14.2.2"
    args.readonly = False
    args.import_path = "./"
    args.export_path = "./"
    args.allow_no_env = False
    args.target_file = "some/file"
    args.setup = DEFAULT_CONFIG_FILE
    return args


def test_config_ok(fm_executor_env):
    args = _mocked_args()

    config = FMRMSConfig(args)
    assert config is not None


def test_missing_project():
    args = _mocked_args()
    args.project = "another_project"

    with pytest.raises(RMSProjectNotFoundError):
        FMRMSConfig(args)


@pytest.mark.parametrize("rms_seed", ["", "a", "123x"])
def test_bad_seed_from_env_var(fm_executor_env, monkeypatch, rms_seed):
    args = _mocked_args()
    monkeypatch.setenv("RMS_SEED", rms_seed)

    with pytest.raises(ValueError, match="'RMS_SEED' environment variable"):
        FMRMSConfig(args)


@pytest.mark.parametrize(
    "iens, expected_result",
    [
        (0, 422851785),
        (1, 422851785),
        (2, 422851785),
    ],
)
def test_single_seed_ok(fm_executor_env, iens, expected_result):
    args = _mocked_args()
    args.iens = iens
    contents = "422851785"
    with open("run_path/RMS_SEED", "w") as f:
        f.write(contents)

    config = FMRMSConfig(args)
    assert config.seed == expected_result


@pytest.mark.parametrize(
    "iens, expected_result",
    [
        (0, 422851785),
        (1, 723121249),
        (2, 132312123),
    ],
)
def test_multi_seed_ok(fm_executor_env, iens, expected_result):
    args = _mocked_args()
    args.iens = iens
    contents = ["3", "422851785", "723121249", "132312123"]
    with open("run_path/random.seeds", "w") as f:
        f.write("\n".join(contents))

    config = FMRMSConfig(args)
    assert config.seed == expected_result


@pytest.mark.parametrize(
    "contents, expected_error",
    [
        ("", r"Single seed file \S+ is empty"),
        ("text", r"Single seed file \S+ contains non-number values"),
        (
            "\n".join(["1000", "1001"]),
            r"Single seed file \S+ contains multiple seed values",
        ),
    ],
)
def test_single_seed_invalid(fm_executor_env, contents, expected_error):
    args = _mocked_args()
    with open("run_path/RMS_SEED", "w") as f:
        f.write(contents)

    with pytest.raises(ValueError, match=expected_error):
        FMRMSConfig(args)


@pytest.mark.parametrize(
    "contents, iens, expected_error",
    [
        ([""], 0, r"Multi seed file \S+ is empty"),
        (["1", "text"], 0, r"Multi seed file \S+ contains non-number values"),
        (
            ["2", "1000"],
            0,
            r"Multi seed file \S+ has an incorrect number count value "
            + "in line 1, expected 1 but found 2",
        ),
        (["0"], 0, r"Multi seed file \S+ has no seed values"),
        (
            ["2", "1000", "1000"],
            0,
            r"Multi seed file \S+ contains non-unique seed values",
        ),
        (
            ["1", "1000"],
            1,
            r"Multi seed file \S+ has too few seed values \(1\) "
            + r"for the needed realization number \(2\)",
        ),
    ],
)
def test_multi_seed_invalid(fm_executor_env, contents, iens, expected_error):
    args = _mocked_args()
    args.iens = iens
    with open("run_path/random.seeds", "w") as f:
        f.write("\n".join(contents))

    with pytest.raises(ValueError, match=expected_error):
        FMRMSConfig(args)



from __future__ import annotations

from ert import (  # type: ignore
    ForwardModelStepDocumentation,
    ForwardModelStepJSON,
    ForwardModelStepPlugin,
    ForwardModelStepValidationError,
    plugin,
)

from runrms.config.fm_rms_config import (
    FMRMSConfig,
    description,
    examples,
)


class Rms(ForwardModelStepPlugin):  # type: ignore
    def __init__(self) -> None:
        super().__init__(
            name="RMS",
            command=[
                "runrms",
                "<RMS_PROJECT>",
                "--batch",
                "<RMS_WORKFLOW>",
                "--iens",
                "<IENS>",
                "--run-path",
                "<RMS_RUNPATH>",
                "--target-file",
                "<RMS_TARGET_FILE>",
                "--import-path",
                "<RMS_IMPORT_PATH>",
                "-v",
                "<RMS_VERSION>",
                "--export-path",
                "<RMS_EXPORT_PATH>",
                "<RMS_OPTS>",
            ],
            default_mapping={
                "<RMS_IMPORT_PATH>": "./",
                "<RMS_EXPORT_PATH>": "./",
                "<RMS_RUNPATH>": "rms/model",
                "<RMS_OPTS>": "",
            },
            target_file="<RMS_TARGET_FILE>",
            exec_env={
                "PYTHONPATH": "<RMS_PYTHONPATH>",
                "PATH_PREFIX": "<RMS_PATH_PREFIX>",
            },
        )

    def validate_pre_realization_run(
        self, fm_step_json: ForwardModelStepJSON
    ) -> ForwardModelStepJSON:
        return fm_step_json

    def validate_pre_experiment(self, fm_step_json: ForwardModelStepJSON) -> None:
        ok, err = FMRMSConfig._pre_experiment_validation()
        if not ok:
            raise ForwardModelStepValidationError(f"FMRMSConfig: {err}")

    @staticmethod
    def documentation() -> ForwardModelStepDocumentation | None:
        return ForwardModelStepDocumentation(
            category="modelling.reservoir",
            source_package="runrms",
            source_function_name="Rms",
            description=description,
            examples=examples,
        )


@plugin(name="runrms")  # type: ignore
def forward_model_configuration() -> dict[str, dict[str, str]]:
    return {
        Rms().name: {
            "RMS_PYTHONPATH": "<RMS_PYTHONPATH>",
            "RMS_PATH_PREFIX": "<RMS_PATH_PREFIX>",
        }
    }


@plugin(name="runrms")  # type: ignore
def installable_forward_model_steps() -> list[ForwardModelStepPlugin]:
    return [Rms]



from __future__ import annotations

from ._rms_config import (
    DEFAULT_CONFIG_FILE,
    RMSConfigNotFoundError,
    RMSExecutableError,
    RMSWrapperError,
)
from ._rms_project import RMSProjectNotFoundError
from .fm_rms_config import FMRMSConfig
from .interactive_rms_config import InteractiveRMSConfig

__all__ = [
    "DEFAULT_CONFIG_FILE",
    "FMRMSConfig",
    "InteractiveRMSConfig",
    "RMSProjectNotFoundError",
    "RMSConfigNotFoundError",
    "RMSWrapperError",
    "RMSExecutableError",
]



from __future__ import annotations

import os
import stat
from pathlib import Path
from textwrap import dedent

import pydantic
import pytest
import yaml

from runrms.__main__ import get_parser
from runrms.config import DEFAULT_CONFIG_FILE
from runrms.config._rms_config import (
    RMSConfig,
    RMSConfigNotFoundError,
    RMSExecutableError,
    RMSVersionError,
    RMSWrapperError,
    _load_site_config,
    _resolve_version,
)
from runrms.config._rms_project import RMSProject
from runrms.config._site_config import SiteConfig
from runrms.config.interactive_rms_config import InteractiveRMSConfig


def test_resolve_version(default_config_file) -> None:
    site_config = SiteConfig.model_validate(default_config_file)
    assert _resolve_version("14.2.1", site_config, None) == "14.2.1"
    assert _resolve_version("14.5.0", site_config, None) == "14.5.0"
    assert _resolve_version(None, site_config, None) == "14.2.2"

    with pytest.raises(
        RMSVersionError, match="RMS version '123.4.5' is not supported."
    ):
        _resolve_version("123.4.5", site_config, None)

    with pytest.raises(RMSVersionError, match="RMS version 'latest' is not supported."):
        _resolve_version("latest", site_config, None)


def test_resolve_version_from_project_master(default_config_file, executor_env):
    site_config = SiteConfig.model_validate(default_config_file)
    rms_project = RMSProject.from_filepath("project")
    version = _resolve_version(None, site_config, rms_project)
    assert version == "14.2.2"

    rms_project.master.version = "14.2.2"
    version = _resolve_version(None, site_config, rms_project)
    assert version == "14.2.2"

    rms_project.master.version = "10.0.0"
    with pytest.raises(
        RMSVersionError,
        match="RMS version '10.0.0' configured in the RMS project is not supported.",
    ):
        _resolve_version(None, site_config, rms_project)


@pytest.mark.parametrize(
    "config_location, raises",
    [
        ("/foo/bar", True),
        (f"{DEFAULT_CONFIG_FILE}.yml", True),
        (DEFAULT_CONFIG_FILE, False),
    ],
)
def test_load_site_config(config_location, raises) -> None:
    if not raises:
        _load_site_config(config_location)
    else:
        with pytest.raises(RMSConfigNotFoundError, match="Unable to locate"):
            _load_site_config(config_location)


def test_valid_config(simple_runrms_config) -> None:
    SiteConfig.model_validate(simple_runrms_config)


def test_invalid_default_version(simple_runrms_config) -> None:
    simple_runrms_config["default"] = "./foo"
    with pytest.raises(pydantic.ValidationError, match="Default RMS version"):
        SiteConfig.model_validate(simple_runrms_config)


def test_init_rmsconfig_default() -> None:
    with open(DEFAULT_CONFIG_FILE, encoding="utf-8") as f:
        config_yml = yaml.safe_load(f)
    config = RMSConfig()
    assert config.version == config_yml["default"]
    assert config.site_config.exe == config_yml["exe"]


def test_init_rmsconfig_default_version(default_config_file) -> None:
    config = RMSConfig()
    assert config.version == default_config_file["default"]
    assert str(config.site_config.exe) == default_config_file["exe"]
    assert (
        default_config_file["versions"][config.version]["env"]["PYTHONPATH"]
        == config.version_config.env.PYTHONPATH
    )


@pytest.mark.parametrize("version", ["14.2.2", "14.5.0"])
def test_init_rmsconfig_given_version(default_config_file, version) -> None:
    config = RMSConfig(version=version)
    assert config.version == version
    assert str(config.site_config.exe) == default_config_file["exe"]
    assert (
        default_config_file["versions"][config.version]["env"]["PYTHONPATH"]
        == config.version_config.env.PYTHONPATH
    )


def test_init_rmsconfig_nondefault_setup(
    tmp_path, simple_runrms_yml, simple_runrms_config
) -> None:
    """Tests that RMSConfig properly handles non-default runrms site configs."""
    runrms_yml = f"{tmp_path}/runrms.yml"
    with open(runrms_yml, "w", encoding="utf-8") as f:
        f.write(simple_runrms_yml("."))

    config = RMSConfig(config_path=runrms_yml)
    assert config.version == "14.2.2"
    assert str(config.site_config.exe) == simple_runrms_config["exe"]
    assert (
        simple_runrms_config["versions"][config.version]["env"]["PYTHONPATH"]
        == config.version_config.env.PYTHONPATH
    )


def test_rmsconfig_get_executable(
    tmp_path,
    monkeypatch,
    simple_runrms_yml,
) -> None:
    """Tests that the wrapper functions correctly."""
    monkeypatch.chdir(tmp_path)
    exe_path = tmp_path / "bin"
    with open("runrms.yml", "w", encoding="utf-8") as f:
        f.write(simple_runrms_yml(exe_path))
    config = RMSConfig(config_path="runrms.yml")
    with pytest.raises(RMSExecutableError, match=f"{exe_path}/rms cannot be found"):
        _ = config.executable

    os.mkdir("bin")
    with open("bin/rms", "w", encoding="utf-8") as f:
        f.write("#!/bin/bash\necho 1")

    with pytest.raises(RMSExecutableError, match=f"{exe_path}/rms cannot be found"):
        _ = config.executable

    os.chmod("bin/rms", stat.S_IEXEC)
    assert config.executable == f"{exe_path}/rms"


def test_rmsconfig_get_wrapper(
    tmp_path,
    monkeypatch,
    simple_runrms_yml,
) -> None:
    """Tests that the wrapper functions correctly."""
    monkeypatch.chdir(tmp_path)
    with open("runrms.yml", "w", encoding="utf-8") as f:
        f.write(simple_runrms_yml(tmp_path))
    config = RMSConfig(config_path="runrms.yml")
    with pytest.raises(RMSWrapperError, match="disable_foo cannot be found"):
        _ = config.wrapper

    disable_foo = Path("disable_foo")
    with open(disable_foo, "w", encoding="utf-8") as f:
        f.write("#!/bin/bash\n")
    disable_foo.chmod(disable_foo.stat().st_mode | stat.S_IEXEC)

    path = os.getenv("PATH", "")
    monkeypatch.setenv("PATH", f"{path}{os.pathsep}{os.getcwd()}")

    assert config.wrapper == "disable_foo"


def test_rmsconfig_get_env(
    tmp_path,
    monkeypatch,
    simple_runrms_yml,
) -> None:
    """Tests that the wrapper functions correctly."""
    monkeypatch.chdir(tmp_path)
    with open("runrms.yml", "w", encoding="utf-8") as f:
        f.write(simple_runrms_yml(tmp_path))
    config = RMSConfig(config_path="runrms.yml")
    assert config.global_env.PATH_PREFIX == "/foo/bin"
    assert config.env.PYTHONPATH == "/foo/bar/site-packages"


@pytest.mark.parametrize(
    "project_version, master_version, expected_version",
    [
        ("14.2.2", "14.2.2", "14.2.2"),
        ("14.5.0", "14.2.2", "14.2.2"),
        ("14.2.1", "V14.2", "14.2.2"),
        ("14.2.2", "V14.2", "14.2.2"),
        ("14.2.1", "V14.2", "14.2.2"),
        ("14.5.0", "V14.5", "14.5.0"),
    ],
)
def test_rmsconfig_with_v14_from_master_resolves_to_latest_patch(
    project_version,
    master_version,
    expected_version,
    tmp_path,
    monkeypatch,
) -> None:
    """Tests that if the .master file does not contain a patch version, as is the case
    in RMS 14.1, it will default to resolving to the latest patch version rather than
    the 0 default version. This test must be updated whenever a new RMS 14.2.x patch
    version is added."""
    monkeypatch.chdir(tmp_path)

    project = f"drogon.{project_version}"
    os.mkdir(project)
    with open(f"{project}/.master", "w", encoding="utf-8") as f:
        f.write(
            dedent(
                f"""
        Begin GEOMATIC file header
        date(1395)                              = 2022.09.08
        time(1395)                              = 10:58:55
        user(1395)                              = user
        release(1395)                           = {master_version}
        operation(1395)                         = Save
        description(1395)                       =
        branch(1395)                            = 14_0
        build(1395)                             = 833
        variant(1395)                           = linux-amd64-gcc_4_8-release
        elements                                = 29
        filetype                                = BINARY
        fileversion                             = 2021.0000
        End GEOMATIC file header
        """
            )
        )

    args = get_parser().parse_args([project])
    config = InteractiveRMSConfig(args)
    assert config.version == expected_version



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Self

from packaging.version import parse as version_parse
from pydantic import BaseModel, Field, model_validator

if TYPE_CHECKING:
    from pathlib import Path


class Env(BaseModel):
    APS_TOOLBOX_PATH: str | None = Field(default=None)
    PYTHONPATH: str
    RMS_PLUGINS_LIBRARY: str
    TCL_LIBRARY: str
    TK_LIBRARY: str


class Version(BaseModel):
    """Information about different RMS versions."""

    env: Env


class GlobalEnv(BaseModel):
    """Top-level environment variables that are set for _all_ RMS versions."""

    PATH_PREFIX: str
    RMS_IPL_ARGS_TO_PYTHON: int = Field(default=1)
    LM_LICENSE_FILE: str | None = Field(default=None)


class SiteConfig(BaseModel):
    """
    Common config class for all RMSConfigs used by runrms
    """

    wrapper: str
    default: str
    exe: str
    interactive_usage_log: Path | None = Field(default=None)
    batch_lm_license_file: str | None = Field(default=None)
    env: GlobalEnv
    versions: dict[str, Version]

    def get_newest_patch_version(self, major: int, minor: int) -> int:
        latest = max(
            version_parse(v) for v in self.versions if v.startswith(f"{major}.{minor}")
        )
        _, _, patch = latest.release
        return patch

    @model_validator(mode="after")
    def default_version_exists_validator(self) -> Self:
        """Validates that the `default` provided actually exists as a key in
        `versions`."""
        try:
            self.versions[self.default]
        except KeyError as e:
            raise ValueError(
                f"Default RMS version {self.default} does not have a corresponding "
                "configuration."
            ) from e
        return self



from __future__ import annotations

from pathlib import Path


def runrms_config_path() -> Path:
    """Returns the absolute path to runrms.yml."""
    return (Path(__file__).parent / "runrms.yml").resolve()



from __future__ import annotations

import argparse
import subprocess
import sys

from .config import FMRMSConfig, InteractiveRMSConfig
from .executor import FMRMSExecutor, InteractiveRMSExecutor

try:
    from .version import __version__
except ImportError:
    __version__ = "0.0.0"


DESCRIPTION = """
This utility runs or opens an RMS project from the command line. In practice it
is a wrapper around the 'rms' command, itself a wrapper that invokes the RMS
application. Note that not all options valid for 'rms' are mapped into this
utility.

This utility can:
- find the RMS version in RMS project files and launch the correct version
- set environment variables, including a PYTHONPATH, for a site configuration
  that extends the vendor distributed Python packages.
- set the path (via environment variables) to site installed RMS plugins.

Example usage:

    $ runrms drogon.rms14.2.2            # If new project: warn and run the RMS default
    $ runrms drogon.rms14.2.2            # Automatically detect version from .master
    $ runrms drogon.rms14.1.3 -v 14.2.2  # Force version 14.2.2, upgrading the project
"""


def _add_fm_arguments(prs: argparse.ArgumentParser) -> None:
    """Command line options specific to the ERT forward model only."""
    fm_prs = prs.add_argument_group(
        title="ERT forward model options",
        description=(
            "These arguments are relevant for the ERT forward model only. "
            "Specifying '--iens' will cause RMS to be run as an RMS forward model. "
            "Typically you should not be providing these options manually."
        ),
    )
    fm_prs.add_argument(
        "--iens",
        dest="iens",
        type=int,
        help="The ERT realization number",
        default=None,
    )
    fm_prs.add_argument(
        "--run-path",
        default="rms/model",
        help="The directory which will be used as current working directory "
        "when ERT is running rms",
    )
    fm_prs.add_argument(
        "--target-file",
        default=None,
        help="Name of file which should be created/updated by rms",
    )
    fm_prs.add_argument(
        "--import-path",
        default="./",
        help="The prefix of all relative paths when rms is importing",
    )
    fm_prs.add_argument(
        "--export-path",
        default="./",
        help="The prefix of all relative paths when rms is exporting",
    )
    fm_prs.add_argument(
        "--allow-no-env",
        action="store_true",
        help="Allow RMS to run without a site configured environment",
    )


def _add_dev_arguments(prs: argparse.ArgumentParser) -> None:
    """Command line options specific to the ERT forward model only."""
    dev_prs = prs.add_argument_group(
        title="Development/testing options",
        description=(
            "These arguments are for development and testing. "
            "Typically you do not need to use these options."
        ),
    )
    dev_prs.add_argument(
        "--test-env",
        "--testpylib",
        dest="testpylib",
        action="store_true",
        help="This option is deprecated. Use '--setup' or 'rmsenv' instead.",
    )
    dev_prs.add_argument(
        "--setup",
        dest="setup",
        type=str,
        help=(
            "Path to the runrms.yml site configuration. "
            "Defaults to the site configuration actually used."
        ),
    )


def get_parser() -> argparse.ArgumentParser:
    """Make a parser for command line arguments and for documentation."""
    prs = argparse.ArgumentParser(
        description=DESCRIPTION,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    prs.add_argument("project", type=str, nargs="?", help="RMS project name")
    prs.add_argument(
        "--debug",
        dest="debug",
        action="store_true",
        help="If you want to run this script in verbose mode",
    )
    prs.add_argument(
        "--project",
        "-p",
        dest="theproject",
        help="Project argument, for backward compaibility (prefer use without this!)",
    )
    prs.add_argument(
        "--dryrun",
        dest="dryrun",
        action="store_true",
        help="Run this script without actually launching RMS",
    )
    prs.add_argument(
        "--version",
        "-v",
        dest="version",
        type=str,
        nargs="?",
        help="The RMS version to run, e.g. 14.2.2.",
    )
    prs.add_argument(
        "--listversions",
        "-l",
        dest="listversions",
        action="store_true",
        help="Use this option to list current RMS versions available. If this option "
        "is set then all other options are disabled.",
    )
    prs.add_argument(
        "--readonly",
        "-r",
        "-readonly",
        dest="readonly",
        action="store_true",
        help="Read only mode (disable save)",
    )
    prs.add_argument(
        "--dpiscaling",
        "-d",
        dest="sdpi",
        default=1.0,
        type=float,
        help="Specify RMS DPI display scaling as a ratio, where 1.0 is no scaling",
    )
    prs.add_argument(
        "--batch",
        "-batch",
        "-w",
        dest="workflow",
        type=str,
        help="Runs project in batch mode with the provided workflow",
    )
    prs.add_argument(
        "--seed",
        "-seed",
        type=int,
        help="The seed to run RMS with. Must be combined with --batch and a project.",
    )
    prs.add_argument(
        "--threads",
        "-threads",
        dest="threads",
        default=1,
        type=int,
        help="The number of threads RMS should use while running.",
    )
    _add_fm_arguments(prs)
    _add_dev_arguments(prs)
    return prs


class UnknownConfigError(Exception):
    """
    Custom error class for unknown config objects
    """


def generate_config(args: argparse.Namespace) -> FMRMSConfig | InteractiveRMSConfig:
    """
    Generate a RMSConfig object based on the given args
    """

    if args.iens is None:
        return InteractiveRMSConfig(args)
    return FMRMSConfig(args)


def generate_executor(
    config: FMRMSConfig | InteractiveRMSConfig,
) -> InteractiveRMSExecutor | FMRMSExecutor:
    """
    Generate a RMSExecutor object based on the given RMSConfig
    """

    if isinstance(config, InteractiveRMSConfig):
        return InteractiveRMSExecutor(config)
    if isinstance(config, FMRMSConfig):
        return FMRMSExecutor(config)

    raise UnknownConfigError(
        f"Unable to generate executor for config of type {type(config)}"
    )


def _validate_args(parsed_args: argparse.Namespace) -> None:
    if parsed_args.seed and not (parsed_args.workflow and parsed_args.project):
        raise argparse.ArgumentError(
            None, "The --seed option must be combined with --batch and a project."
        )


def main(args: list[str] | None = None) -> int:
    """
    Launch rms
    """
    if args is None:
        args = sys.argv[1:]
    args = [x for x in args if x]
    parsed_args = get_parser().parse_args(args)

    _validate_args(parsed_args)

    config = generate_config(parsed_args)

    if parsed_args.listversions:
        subprocess.run(args=[str(config.executable), "-v"], check=False)
        return 0
    if parsed_args.testpylib:
        print(
            "The '--testpylib'/'--test-env' options are deprecated due to a change in"
            " the way that RMS Python environments are deployed. Use '--setup' or an "
            "rmsenv instead."
        )
        return 0

    executor = generate_executor(config)
    return executor.run()



from ._rms_executor import RMSRuntimeError
from .fm_rms_executor import FMRMSExecutor
from .interactive_rms_executor import InteractiveRMSExecutor

__all__ = [
    "FMRMSExecutor",
    "InteractiveRMSExecutor",
    "RMSRuntimeError",
]



import json
import re
import shutil
import subprocess

import pytest

try:
    from ert.plugins.plugin_manager import ErtPluginManager
except ImportError:
    pytest.skip(
        "Not testing ERT plugins when ERT is not installed", allow_module_level=True
    )

import runrms._forward_model as ert_plugins
from runrms._forward_model import Rms

EXPECTED_JOBS = {"RMS"}


@pytest.mark.requires_ert
def test_that_installable_fm_steps_work_as_plugins():
    """Test that the forward models are included as ERT plugin."""
    fms = ErtPluginManager(plugins=[ert_plugins]).forward_model_steps

    assert Rms in fms
    assert len(fms) == len(EXPECTED_JOBS)


@pytest.mark.requires_ert
def test_fm_plugin_implementations():
    """Test hook implementation."""
    ert_pm = ErtPluginManager(plugins=[ert_plugins])

    installable_fm_step_jobs = [fms().name for fms in ert_pm.forward_model_steps]
    assert set(installable_fm_step_jobs) == set(EXPECTED_JOBS)

    installable_workflow_jobs = ert_pm.get_installable_workflow_jobs()
    assert len(installable_workflow_jobs) == 0


@pytest.mark.requires_ert
@pytest.mark.integration
def test_fm_plugin_executables():
    """Test executables in the configured ert forward models."""
    ert_pm = ErtPluginManager(plugins=[ert_plugins])
    for fm_step in ert_pm.forward_model_steps:
        assert shutil.which(fm_step().executable)


@pytest.mark.requires_ert
def test_fm_plugin_docs():
    """For each installed forward model, we require the associated
    description and example string to be nonempty,
    and the category to be as expected"""

    ert_pm = ErtPluginManager([ert_plugins])
    for fm_step in ert_pm.forward_model_steps:
        docs = fm_step.documentation()
        assert docs.description is not None
        assert docs.examples is not None
        assert "RMS" in docs.description
        assert "FORWARD_MODEL RMS" in docs.examples
        assert docs.category == "modelling.reservoir"


@pytest.mark.requires_ert
def test_rms_forward_model_ok(tmp_path, monkeypatch, fmu_snakeoil_project):
    """Test that when running ert with the given configuration file,
    the rms forward model runs rms with the given arguments"""
    monkeypatch.chdir(tmp_path / "ert/model")

    subprocess.run(["ert", "test_run", "snakeoil.ert", "--verbose"], check=False)

    with open(tmp_path / "scratch/user/snakeoil/realization-0/iter-0/jobs.json") as f:
        jobs_json = json.load(f)
        assert jobs_json["config_file"] == "snakeoil.ert"
        assert jobs_json["jobList"][0]["executable"] == "runrms"
        assert jobs_json["jobList"][0]["name"] == "RMS"
        assert jobs_json["jobList"][0]["argList"][4] == "0"
        assert jobs_json["jobList"][0]["argList"][12] == "14.2.2"


@pytest.mark.requires_ert
def test_rms_forward_model_seed_invalid(
    tmp_path, monkeypatch, fmu_snakeoil_project, create_multi_seed_file
):
    """Test that when the seed file used by RMS has an invalid format,
    the validation leads to an aborted run"""
    monkeypatch.chdir(tmp_path / "ert/model")
    create_multi_seed_file("text\n")

    output = subprocess.run(
        ["ert", "test_run", "snakeoil.ert"], capture_output=True, text=True, check=False
    )

    assert re.search(
        "Forward model step pre-experiment validation failed: "
        + r"FMRMSConfig: Multi seed file \S+ contains non-number values",
        output.stderr,
    )



from __future__ import annotations

import logging

logger = logging.getLogger(__name__)


class BColors:
    # pylint: disable=too-few-public-methods
    # local class for ANSI term color commands

    HEADER = "\033[93;42m"
    OKBLUE = "\033[94m"
    OKGREEN = "\033[92m"
    WARN = "\033[93;43m"
    ERROR = "\033[93;41m"
    CRITICAL = "\033[1;91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"


def xwarn(mystring: str) -> None:
    """Print a warning with colors."""
    print(BColors.WARN, mystring, BColors.ENDC)


def xalert(mystring: str) -> None:
    """Print an alert warning in an appropriate color."""
    print(BColors.ERROR, mystring, BColors.ENDC)


def xcritical(mystring: str) -> None:
    """Print an critical error in an appropriate color."""
    print(BColors.CRITICAL, mystring, BColors.ENDC)



from __future__ import annotations

import logging
import os
import random
from pathlib import Path
from typing import TYPE_CHECKING

from runrms.executor import RMSRuntimeError

from ._rms_config import RMSConfig

if TYPE_CHECKING:
    import argparse

    from ._rms_project import RMSProject

logger = logging.getLogger(__name__)

description = """
Forward model for running a given workflow in an existing RMS-project.

The forward model requires explicit knowledge of the version used to
produce the RMS project loaded. The Python environment is adapted to
work with Python inside RMS.

As part of the configuration there can be a file holding a list of random seeds:
``ert/input/distributions/random.seeds``. This file needs to have a specific format,
containing one number per line, with each number being unique within the file. In
addition, the first line of the file needs to have a count of the numbers in the file.
If there are 1000 unique numbers in the file, the first line will show 1000, and the
total number of lines in the file will be 1001.
"""

examples = """
Running the forward model
#########################

RMS is usually incorporated in ERT configurations using statements like

.. code-block:: bash

    DEFINE  <RMS_PROJECT>         reek.rms11.0.1
    DEFINE  <RMS_VERSION>         11.0.1
    DEFINE  <RMS_WORKFLOW_NAME>   MAIN_WORKFLOW
    FORWARD_MODEL RMS(<IENS>=<IENS>, <RMS_VERSION>=<RMS_VERSION>, <RMS_PROJECT>=<CONFIG_PATH>/../../rms/model/<RMS_NAME>)
"""  # noqa

category = "modelling.reservoir"


class FMRMSConfig(RMSConfig):
    """A class which holds the nessecary configuration for executing
    runrms as a forward model.
    """

    project: RMSProject

    _single_seed_file = "RMS_SEED"
    _multi_seed_file = "random.seeds"
    _max_seed = 2146483648
    _seed_factor = 7907

    def __init__(self, args: argparse.Namespace) -> None:
        if not args.project:
            raise RMSRuntimeError(
                "A project must be specified to run the RMS forward model."
            )
        super().__init__(
            config_path=args.setup, version=args.version, project=args.project
        )
        self._iens = args.iens
        self._threads = args.threads
        self._workflow = args.workflow

        self._run_path = Path(args.run_path)
        self._import_path = Path(args.import_path)
        self._export_path = Path(args.export_path)
        self._allow_no_env = args.allow_no_env

        self._seed = (
            self._get_seed_from_env()
            if "RMS_SEED" in os.environ
            else self._read_seed_from_file(self._iens)
        ) % self._max_seed

        self._target_file = None
        self._target_file_mtime = None
        if args.target_file:
            self._target_file = Path(
                args.target_file
                if os.path.isabs(args.target_file)
                else os.path.join(os.getcwd(), args.target_file)
            )
            self._target_file_mtime = (
                os.path.getmtime(args.target_file)
                if os.path.isabs(args.target_file)
                else None
            )

    @property
    def allow_no_env(self) -> Path:
        return self._allow_no_env

    @property
    def run_path(self) -> Path:
        return self._run_path

    @property
    def import_path(self) -> Path:
        return self._import_path

    @property
    def export_path(self) -> Path:
        return self._export_path

    @property
    def target_file(self) -> Path | None:
        return self._target_file

    @property
    def target_file_mtime(self) -> float | None:
        return self._target_file_mtime

    @property
    def seed(self) -> int:
        return self._seed

    @property
    def threads(self) -> int:
        return self._threads

    @property
    def workflow(self) -> str:
        return self._workflow

    def _get_seed_from_env(self) -> int:
        try:
            env_seed = os.environ.get("RMS_SEED", "")
            seed = int(env_seed)
        except ValueError as e:
            raise ValueError(
                f"The 'RMS_SEED' environment variable is set to {env_seed} which "
                "cannot be converted into an integer."
            ) from e

        for _ in range(self._iens):
            seed *= self._seed_factor
        return seed

    def _read_seed_from_file(self, iens: int) -> int:
        single_seed_file = self.run_path / self._single_seed_file
        multi_seed_file = self.run_path / self._multi_seed_file

        if single_seed_file.exists():
            # Using existing single seed file
            with open(single_seed_file) as file_handle:
                seed_list = [line.rstrip() for line in file_handle]
                self._validate_seed_source(seed_list, single_seed_file, False, iens)
                seed = int(seed_list[0])
        elif multi_seed_file.exists():
            with open(multi_seed_file) as file_handle:
                seed_list = [line.rstrip() for line in file_handle]
                self._validate_seed_source(seed_list, multi_seed_file, True, iens)
                seed = int(seed_list[self._iens + 1])
        else:
            random.seed()
            seed = random.randint(0, FMRMSConfig._max_seed)
        return seed

    @staticmethod
    def _pre_experiment_validation(
        num_realizations: int | None = None,
    ) -> tuple[bool, ValueError | None]:
        seed_path = Path(os.getcwd(), "../input/distributions")
        single_seed_file = seed_path / FMRMSConfig._single_seed_file
        multi_seed_file = seed_path / FMRMSConfig._multi_seed_file
        iens_max = num_realizations - 1 if num_realizations is not None else None

        try:
            if single_seed_file.exists():
                with open(single_seed_file) as file_handle:
                    seed_list = [line.rstrip() for line in file_handle]
                    FMRMSConfig._validate_seed_source(
                        seed_list, single_seed_file, False, iens_max
                    )
            elif multi_seed_file.exists():
                with open(multi_seed_file) as file_handle:
                    seed_list = [line.rstrip() for line in file_handle]
                    FMRMSConfig._validate_seed_source(
                        seed_list, multi_seed_file, True, iens_max
                    )
        except ValueError as err:
            return False, err
        return True, None

    @staticmethod
    def _validate_seed_source(
        lines: list[str], filename: Path, is_multi: bool, iens_max: int | None = None
    ) -> None:
        file_desc = "Multi seed file" if is_multi else "Single seed file"
        file_desc += f" {filename.absolute()}"
        single_format_desc = "The file must contain one number"
        multi_format_desc = (
            "The file contents must be unique numbers, one per line. "
            "The first line must have a count of the total numbers in "
            "the file, excluding the count value itself"
        )
        format_desc = multi_format_desc if is_multi else single_format_desc

        line_count = len(lines)
        if line_count == 0:
            raise ValueError(f"{file_desc} is empty. {format_desc}")
        for line in lines:
            if not line.isdigit():
                raise ValueError(
                    f"{file_desc} contains non-number values. {format_desc}"
                )
        if not is_multi:
            if line_count > 1:
                raise ValueError(
                    f"{file_desc} contains multiple seed values. {format_desc}"
                )
            return

        number_count = int(lines[0])
        numbers = lines[1:]
        if number_count != len(numbers):
            raise ValueError(
                f"{file_desc} has an incorrect number count value in line 1, "
                + f"expected {len(numbers)} but found {number_count}. {format_desc}"
            )
        if number_count == 0:
            raise ValueError(f"{file_desc} has no seed values. {format_desc}")
        if len(numbers) != len(set(numbers)):
            raise ValueError(
                f"{file_desc} contains non-unique seed values. {format_desc}"
            )
        if iens_max is not None and number_count <= iens_max:
            raise ValueError(
                f"{file_desc} has too few seed values ({number_count}) "
                + f"for the needed realization number ({iens_max + 1})"
            )



import sys
from argparse import ArgumentError

import pytest

from runrms.__main__ import generate_config, get_parser, main


def test_empty_invocation(executor_env, patch_argv) -> None:
    patch_argv(["--setup", "runrms.yml"])
    main()


def test_invalid_batch_invocations(executor_env, patch_argv) -> None:
    patch_argv(["--setup", "runrms.yml", "--seed", "123"])
    with pytest.raises(ArgumentError, match="must be combined with --batch"):
        main()

    patch_argv(["--setup", "runrms.yml", "--seed", "123", "--batch", "a"])
    with pytest.raises(ArgumentError, match="must be combined with --batch"):
        main()

    patch_argv(["--setup", "runrms.yml", "--seed", "123", "a"])
    with pytest.raises(ArgumentError, match="must be combined with --batch"):
        main()

    patch_argv(["b", "--setup", "runrms.yml", "--seed", "123", "--batch", "a"])
    with pytest.raises(OSError, match="does not exist as a directory"):
        main()

    patch_argv(["project", "--setup", "runrms.yml", "-w", "a", "b"])
    with pytest.raises(SystemExit):
        main()

    patch_argv(["project", "--setup", "runrms.yml", "--batch", "a"])
    args = get_parser().parse_args(sys.argv[1:])
    config = generate_config(args)
    assert config.workflow == "a"



from __future__ import annotations

import logging
import os
import shutil
from importlib import metadata as importlib_metadata
from pathlib import Path
from typing import Final

import yaml
from packaging.version import parse as version_parse

from ._rms_project import RMSProject
from ._site_config import Env, GlobalEnv, SiteConfig, Version

logger = logging.getLogger(__name__)


DEFAULT_CONFIG_FILE: Final = os.path.join(os.path.dirname(__file__), "runrms.yml")
DEFAULT_VERSION: Final = "DEFAULT"


class RMSConfigError(ValueError):
    """Raised when the configuration file is errorneous."""


class RMSConfigNotFoundError(FileNotFoundError):
    """Raised when attempting to open a site configuration that does not exist."""


class RMSExecutableError(OSError):
    """Raised when the RMS executable cannot be executed, either because it's
    not found or because of invalid user access to it."""


class RMSWrapperError(FileNotFoundError):
    """Raised when the RMS wrapper cannot be found."""


class RMSVersionError(ValueError):
    """Raised when the given rms version does not exist."""


def _detect_os() -> str:
    """Detect operating system string in runtime. Use default if not found."""

    default_os_version = "x86_64_RH_8"
    release_file = Path("/etc/redhat-release")
    if release_file.is_file():
        with open(release_file, encoding="utf-8") as f:
            tokens = f.read().split()
            for t in tokens:
                if "." in t:
                    major = t.split(".")[0]
                    osver = f"x86_64_RH_{major}"
                    logger.debug(f"RHEL version {osver} found in {release_file}")
                    return osver
        raise ValueError("Could not detect RHEL version")
    return default_os_version


def _load_site_config(site_config_file: str) -> SiteConfig:
    if not os.path.exists(site_config_file):
        raise RMSConfigNotFoundError(
            "Unable to locate config file for rms\n"
            f"{site_config_file} does not exist!"
        )
    with open(site_config_file) as f:
        config = yaml.safe_load(f)
    return SiteConfig.model_validate(config)


def _resolve_version(
    version: str | None, site_config: SiteConfig, rms_project: RMSProject | None
) -> str:
    if version:
        if version in site_config.versions:
            return version
        raise RMSVersionError(
            f"RMS version '{version}' is not supported. "
            "To see the supported versions, run `rms -l` or `runrms -l`."
        )
    if rms_project:
        if rms_project.master.version in site_config.versions:
            master_version = version_parse(rms_project.master.version)
            major, minor, patch = master_version.release

            # Handle RMS 14 specially as it stores no patch version internally.
            if major == 14:
                newest_patch = site_config.get_newest_patch_version(major, minor)
                return f"{major}.{minor}.{newest_patch}"

            return rms_project.master.version

        raise RMSVersionError(
            f"RMS version '{rms_project.master.version}' "
            "configured in the RMS project is not supported."
            "To see the supported versions, run `rms -l` or `runrms -l`."
        )

    return site_config.default


class RMSConfig:
    """
    Common config class for all RMSConfigs used by runrms
    """

    def __init__(
        self,
        *,
        config_path: str | None = None,
        version: str | None = None,
        project: str | None = None,
    ) -> None:
        super().__init__()
        self._osver = _detect_os()

        self._site_config_file = self._set_config_file(config_path)
        self._site_config = _load_site_config(self._site_config_file)
        self._project = RMSProject.from_filepath(project) if project else None

        self._version_given = version
        self._version = _resolve_version(version, self._site_config, self._project)
        self._version_config = self._site_config.versions[self._version]

    def _set_config_file(self, config_path: str | None) -> str:
        """Determines which configuration file to use.

        Starts with the one included within this package, then looks for any exposed by
        a runrms entry point, and finally prefers one given by '--setup'."""
        config_file = DEFAULT_CONFIG_FILE

        entry_points = importlib_metadata.entry_points()
        # Python 3.12 does not implement __iter__ on this object.
        selections = entry_points.select(group="runrms", name="config_path")
        if selections:
            runrms_config_path, *_ = selections
            config_file = runrms_config_path.load()()

        # Override if given from --setup
        if config_path:
            config_file = config_path

        return str(config_file)

    @property
    def osver(self) -> str:
        return self._osver

    @property
    def site_config_file(self) -> str:
        return self._site_config_file

    @property
    def site_config(self) -> SiteConfig:
        return self._site_config

    @property
    def project(self) -> RMSProject | None:
        return self._project

    @property
    def version_given(self) -> str | None:
        return self._version_given

    @property
    def version(self) -> str:
        return self._version

    @property
    def version_config(self) -> Version:
        return self._version_config

    @property
    def executable(self) -> str:
        """RMS executable, assert if permissions are correct"""
        exe = self._site_config.exe
        if shutil.which(exe) is None:
            raise RMSExecutableError(f"The executable: {exe} cannot be found")
        if not os.access(exe, os.X_OK):
            raise RMSExecutableError(
                f"The executable: {exe} cannot be run (invalid access)"
            )
        return exe

    @property
    def wrapper(self) -> str:
        """wrapper executable, assert if permissions are correct"""
        exe = self._site_config.wrapper
        if shutil.which(exe) is None:
            raise RMSWrapperError(f"The executable: {exe} cannot be found")
        return exe

    @property
    def global_env(self) -> GlobalEnv:
        return self._site_config.env

    @property
    def env(self) -> Env:
        """Environment given by the config file"""
        return self._version_config.env

    @property
    def threads(self) -> int:
        """Number of threads to use in RMS. Defaults to one, the recommended number in
        batch mode."""
        return 1



from __future__ import annotations

import glob
import json
import os
import subprocess
import sys
import time
from contextlib import contextmanager
from textwrap import dedent
from typing import TYPE_CHECKING

from ._rms_executor import RMSExecutor, RMSRuntimeError

if TYPE_CHECKING:
    from collections.abc import Generator
    from pathlib import Path

    from runrms.config.fm_rms_config import FMRMSConfig


@contextmanager
def pushd(path: str | Path) -> Generator[None, None, None]:
    """pushd functionality"""
    cwd_ = os.getcwd()
    try:
        os.chdir(path)
        yield
    finally:
        os.chdir(cwd_)


class FMRMSExecutor(RMSExecutor):
    """
    Class for executing runrms as forward model job
    """

    config: FMRMSConfig

    def _exec_rms(self) -> int:
        """Execute RMS with given environment"""
        args = self.pre_rms_args()
        args += [
            str(self.config.executable),
            "-project",
            str(self.config.project.path),
            "-seed",
            str(self.config.seed),
            "-readonly",
            "-nomesa",
            "-export_path",
            str(self.config.export_path),
            "-import_path",
            str(self.config.import_path),
            "-batch",
            str(self.config.workflow),
        ]

        if self.config.version:
            args += ["-v", str(self.config.version)]

        if self.config.threads:
            args += ["-threads", str(self.config.threads)]

        comp_process = subprocess.run(args=args, check=False)
        return comp_process.returncode

    def _add_env_from_json(self) -> None:
        # This function is to be removed/changed
        # PYTHONPATH and PATH_PREFIX that could come from this json
        # should be explicitly handled. These will arrive as
        # RMS_PYTHONPATH and RMS_PATH_PREFIX in the existing environment(?)
        self_exe, _ = os.path.splitext(os.path.basename(sys.argv[0]))
        exec_env_file = f"{self_exe}_exec_env.json"

        user_env = {}
        if os.path.isfile(exec_env_file):
            with open(exec_env_file, encoding="utf-8") as f:
                user_env = json.load(f)

        for key in set(self._exec_env.keys()) | set(user_env.keys()):
            if user_env.get(key):
                self._update_exec_env(key, str(user_env.get(key)), "json")

    def print_failure(self, exit_status: int) -> None:
        run_path = self.config.run_path.resolve()
        # Reverse sort so workflow.log is (probably) first and
        # YYYYMMDD-HHMMSS-XXXXX-RMS.log files are (probably) last
        log_files = sorted(glob.glob(f"{run_path}/*.log"), reverse=True)

        if exit_status == 137:
            # When the OOM-killer strikes, the RMS process (or maybe one of its
            # subprocesses) will get a SIGKILL (9) signal, which is often reported
            # as 128+9=137.
            fail_msg = dedent(
                f"""
        The RMS run failed with exit status: {exit_status}.

        This often means that the compute node ran out of memory and RMS or one of its
        subprocesses was terminated.
                """
            )

        elif not log_files:
            fail_msg = dedent(
                f"""
        The RMS run failed with exit status: {exit_status} and no log files were
        found in:

        * {run_path}

        This may mean that the compute node ran out of memory and RMS or one of its
        subprocesses was terminated, or that some other error has occurred.
                """
            )

        else:
            fail_msg = dedent(
                f"""
        The RMS run failed with exit status: {exit_status}. Typically this means a
        job in an RMS workflow has failed.
                """
            )

        if log_files:
            fail_msg += dedent(
                """
        For more details try checking these log files:

        * RMS.stderr.NN and RMS.stdout.NN
        * rms/model/workflow.log
        * Other named log files in rms/model, e.g. workflow_sim2seis.log
        * rms/model/YYYYMMDD-HHMMSS-XXXXX-RMS.log corresponding to your run

        The following log files were found in this realization's run path:

                """
            )
            fail_msg += "\n".join([f"* {f}" for f in log_files])

        print(fail_msg, file=sys.stderr)

    def run(self) -> int:
        """Main executor entry point."""
        if not os.path.exists(self.config.run_path):
            os.makedirs(self.config.run_path)

        if (
            self.config._version_given != self.config.version
            and not self.config.allow_no_env
        ):
            raise RMSRuntimeError(
                "RMS environment not specified for version: "
                f"{self.config._version_given}"
            )

        self._initialize_exec_env_from_config()
        if (license_file := self.config.site_config.batch_lm_license_file) is not None:
            self._update_exec_env("LM_LICENSE_FILE", license_file, "config")
        self._add_env_from_json()

        with pushd(self.config.run_path):
            now = time.strftime("%d-%m-%Y %H:%M:%S", time.localtime(time.time()))
            with open("RMS_SEED_USED", "a+", encoding="utf-8") as filehandle:
                filehandle.write(f"{now} ... {self.config.seed}\n")

            if not os.path.exists(self.config.export_path):
                os.makedirs(self.config.export_path)

            if not os.path.exists(self.config.import_path):
                os.makedirs(self.config.import_path)

            exit_status = self._exec_rms()

        if exit_status != 0:
            self.print_failure(exit_status)
            return exit_status

        if self.config.target_file is None:
            return exit_status

        if not os.path.isfile(self.config.target_file):
            raise RMSRuntimeError(
                "The RMS run did not produce the expected file: "
                f"{self.config.target_file}"
            )

        if self.config.target_file_mtime is None:
            return exit_status

        if os.path.getmtime(self.config.target_file) == self.config.target_file_mtime:
            raise RMSRuntimeError(
                f"The target file: {self.config.target_file} is unmodified - "
                "interpreted as failure"
            )
        return exit_status



from __future__ import annotations

import subprocess
import sys
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pathlib import Path

    from pytest import MonkeyPatch


def test_entry_point(
    tmp_path: Path, monkeypatch: MonkeyPatch, source_root: Path
) -> None:
    venv_path = tmp_path / "venv"
    rmsconfig = source_root / "examples/rmsconfig"

    pyver = f"{sys.version_info.major}.{sys.version_info.minor}"
    make_venv = subprocess.run(
        f"python -m venv {venv_path};"
        f"source {venv_path}/bin/activate;"
        f"pip install -U pip; pip install {source_root};",
        shell=True,
        executable="/bin/bash",
        check=False,
    )
    assert make_venv.returncode == 0

    get_version_cmd = (
        f"source {venv_path}/bin/activate;"
        'python -c "from runrms.config._rms_config import RMSConfig;'
        'print(RMSConfig()._site_config_file)";'
    )

    runrms_default_config = subprocess.run(
        get_version_cmd,
        capture_output=True,
        text=True,
        shell=True,
        executable="/bin/bash",
        check=False,
    )
    assert runrms_default_config.returncode == 0
    # It may end up in lib64/ over lib/
    assert runrms_default_config.stdout.rstrip() in [
        f"{venv_path}/lib/python{pyver}/site-packages/runrms/config/runrms.yml",
        f"{venv_path}/lib64/python{pyver}/site-packages/runrms/config/runrms.yml",
    ]

    install_rmsconfig = subprocess.run(
        f"source {venv_path}/bin/activate; pip install {rmsconfig};",
        shell=True,
        executable="/bin/bash",
        check=False,
    )
    assert install_rmsconfig.returncode == 0

    check_config_file = subprocess.run(
        get_version_cmd,
        capture_output=True,
        text=True,
        shell=True,
        executable="/bin/bash",
        check=False,
    )
    assert check_config_file.returncode == 0
    assert check_config_file.stdout.rstrip() in [
        f"{venv_path}/lib/python{pyver}/site-packages/rmsconfig/runrms.yml",
        f"{venv_path}/lib64/python{pyver}/site-packages/rmsconfig/runrms.yml",
    ]

    check_dryrun = subprocess.run(
        f"source {venv_path}/bin/activate; runrms -v 19.0.0;",
        capture_output=True,
        text=True,
        shell=True,
        executable="/bin/bash",
        check=False,
    )
    # Command will fail, but assert against the error message
    assert check_dryrun.returncode == 1
    # The example configured exe (which throws this error) differs from the default
    # configuration file.
    assert "/opt/example/rms/ cannot be found" in check_dryrun.stderr



from __future__ import annotations

import os
from abc import ABC, abstractmethod
from copy import deepcopy
from typing import TYPE_CHECKING, Literal

if TYPE_CHECKING:
    from runrms.config import FMRMSConfig, InteractiveRMSConfig


class RMSRuntimeError(Exception):
    """
    Custom error for run-time errors
    """


class RMSExecutor(ABC):
    """
    Executor class which should be used by all runrms executors
    """

    def __init__(self, config: InteractiveRMSConfig | FMRMSConfig) -> None:
        self._config = config

        # We want to append to or carry forward the following env vars
        pre_env = {}
        pre_env["PYTHONPATH"] = os.environ.get("PYTHONPATH", None)
        pre_env["RMS_PLUGINS_LIBRARY"] = os.environ.get("RMS_PLUGINS_LIBRARY", None)
        self._pre_env = {k: v for k, v in pre_env.items() if v}
        self._exec_env: dict[str, str]

        super().__init__()

    @property
    def config(self) -> InteractiveRMSConfig | FMRMSConfig:
        return self._config

    @property
    def pre_env(self) -> dict[str, str]:
        """Returns a dict containing a few specific environment variables we will
        carry forward from before execution."""
        return deepcopy(self._pre_env)

    def _initialize_exec_env_from_config(self) -> None:
        """Initializes the environment variables for the
        execution environment of RMS."""
        self._exec_env = self.pre_env
        config_env = self._config_env()
        for key, val in config_env.items():
            self._update_exec_env(key, val, "config")

    def _update_exec_env(
        self, key: str, val: str, val_origin: Literal["config", "json", "test_config"]
    ) -> None:
        """Updates the environment variable with name `key` in the
        execution environment of RMS with the value `val`.
        Depending on the environment variable to update and the origin of `val`,
        the entry will be updated in different ways. If the key does not match any
        of the keys that needs special handling, the default case will add the value
        to the front of the existing search path.
        """
        # TODO: Replace if/elif pattern below with a match case
        # when all RMS versions support Python 3.10
        if key == "RMS_PLUGINS_LIBRARY":
            if val_origin == "json":
                self._exec_env[key] = val
            elif val_origin == "config":
                if self._exec_env.get(key):
                    return
                self._exec_env[key] = val
            elif val_origin == "test_config" and self._exec_env.get(key):
                self._exec_env[key] = val
                return
        elif key == "LM_LICENSE_FILE":
            self._exec_env[key] = val
        elif key == "QT_SCALE_FACTOR":
            self._exec_env[key] = str(val)
        else:
            # assert this is a PATH thing...
            if key in self._exec_env:
                self._exec_env[key] = f"{val}{os.pathsep}{self._exec_env[key]}"
            else:
                self._exec_env[key] = str(val)
            if not self._exec_env[key].strip():
                self._exec_env.pop(key)

    def _config_env(self) -> dict[str, str]:
        """Returns a dict containing the key, value environment variable pairs from the
        configuration. This merges the top-level global configuration for all RMS
        versions as well as the specific RMS version environment variables. The default
        behavior is to overwrite the global variable if a variable of the same name
        exists in the version configuration."""
        config_env = {
            k: str(v) for k, v in vars(self.config.global_env).items() if v is not None
        }
        version_env = {
            k: str(v) for k, v in vars(self.config.env).items() if v is not None
        }
        # Overwrite the global env if there are conflicts.
        config_env.update(version_env)
        return config_env

    def pre_rms_args(self) -> list[str]:
        """The rms exec environement needs to be injected between executing the
        wrapper and launching rms. PATH_PREFIX must be set in advance."""
        prefix_path = self._exec_env.pop("PATH_PREFIX", "")
        env_args = ["env", *(f"{key}={value}" for key, value in self._exec_env.items())]
        return (
            ["env", f"PATH_PREFIX={prefix_path}", self.config.wrapper] + env_args
            if self.config.wrapper is not None
            else env_args
        )

    @abstractmethod
    def run(self) -> int:
        """Main executor function for running rms"""
        raise NotImplementedError



import shutil
from pathlib import Path

import pytest

from runrms.config._rms_project import (
    RMSProject,
    RMSProjectNotFoundError,
    _parse_master_file_header,
    _sanitize_version,
)


@pytest.mark.parametrize(
    "given, expected",
    [
        ("13", "13.0.0"),
        ("13.1", "13.1.0"),
        ("V14", "14.0.0"),
        ("v14.1.3", "14.1.3"),
        ("14.2b", "14.2b"),
    ],
)
def test_sanitize_version(given: str, expected: str) -> None:
    assert _sanitize_version(given) == expected


def test_make_rmsproject_with_nonexistent_project_dir(tmp_path, monkeypatch) -> None:
    monkeypatch.chdir(tmp_path)
    with pytest.raises(RMSProjectNotFoundError, match="does not exist as a directory"):
        RMSProject.from_filepath("notreal")


def test_make_rmsproject_with_project_dir_as_file(tmp_path, monkeypatch) -> None:
    monkeypatch.chdir(tmp_path)
    Path("notreal").touch()
    with pytest.raises(RMSProjectNotFoundError, match="does not exist as a directory"):
        RMSProject.from_filepath("notreal")


def test_make_rmsproject_without_master_file(tmp_path, monkeypatch) -> None:
    monkeypatch.chdir(tmp_path)
    Path("notreal").mkdir()
    with pytest.raises(FileNotFoundError, match="RMS project .master file not found"):
        RMSProject.from_filepath("notreal")


def test_rmsproject_parses_correct_master(source_root) -> None:
    drogon_master = source_root / "tests/testdata/rms/drogon.rms13.0.3/.master"
    master = _parse_master_file_header(drogon_master)
    assert master.version == "13.0.3"
    assert master.user == "jriv"
    assert master.time == "10:58:55"
    assert master.date == "2022.09.08"
    assert master.variant == "linux-amd64-gcc_4_8-release"
    assert master.fileversion == "2021.0000"


def test_make_rmsproject_without_lock(source_root) -> None:
    drogon_master = source_root / "tests/testdata/rms/drogon.rms13.0.3"
    rmsproject = RMSProject.from_filepath(drogon_master)
    assert rmsproject.path == drogon_master
    assert rmsproject.name == "drogon.rms13.0.3"
    assert rmsproject.locked is False
    assert rmsproject.lockfile is None
    assert rmsproject.master.version == "13.0.3"
    assert rmsproject.master.user == "jriv"
    assert rmsproject.master.time == "10:58:55"
    assert rmsproject.master.date == "2022.09.08"
    assert rmsproject.master.variant == "linux-amd64-gcc_4_8-release"
    assert rmsproject.master.fileversion == "2021.0000"


def test_make_rmsproject_with_lock(tmp_path, monkeypatch, source_root) -> None:
    test_path = tmp_path / "drogon.rms13.0.3"
    drogon_master = source_root / "tests/testdata/rms/drogon.rms13.0.3"
    shutil.copytree(drogon_master, test_path)

    monkeypatch.chdir(test_path)
    lock_txt = (
        "Locked by abc on s6.st.so.no with process id (pid) 123 at 2037.03.14 09:00:00"
    )
    Path("project_lock_file").write_text(lock_txt)
    rmsproject = RMSProject.from_filepath(test_path)
    assert rmsproject.path == test_path
    assert rmsproject.name == "drogon.rms13.0.3"
    assert rmsproject.locked is True
    assert rmsproject.lockfile == lock_txt



from __future__ import annotations

import datetime
import getpass
import logging
import platform
import shutil
import subprocess
import time
from typing import TYPE_CHECKING

from runrms._utils import (
    BColors,
    xalert,
    xwarn,
)
from runrms.version import __version__

from ._rms_executor import RMSExecutor

if TYPE_CHECKING:
    from runrms.config.interactive_rms_config import (
        InteractiveRMSConfig,
    )

logger = logging.getLogger(__name__)


class InteractiveRMSExecutor(RMSExecutor):
    """
    Class for executing runrms in interactive mode
    """

    config: InteractiveRMSConfig

    def _handle_locked_project(self) -> None:
        """Do action if project is locked."""
        xwarn(
            "NB! Opening a locked RMS project (you have 5 seconds to press "
            "Ctrl-C to abort)"
        )
        for sec in range(5, 0, -1):
            time.sleep(1)
            print(f"... {sec}")

    def _exec_rms(self) -> int:
        """Launch RMS with correct pythonpath, pluginspath etc."""
        self._initialize_exec_env_from_config()
        self._update_exec_env(
            "QT_SCALE_FACTOR", str(self.config._dpi_scaling), "config"
        )
        pre_args = self.pre_rms_args()

        args = [
            str(self.config.executable),
            "-v",
            str(self.config.version),
        ]
        if self.config.readonly:
            args.append("-readonly")

        if self.config.workflow:
            args += ["-batch", str(self.config.workflow)]

        if self.config.project:
            args += ["-project", str(self.config.project.path)]

        self.command = " ".join(args)
        print(BColors.BOLD, f"\nRunning: {self.command}\n", BColors.ENDC)
        print("=" * shutil.get_terminal_size((132, 20)).columns)

        if self.config.project and self.config.project.locked:
            self._handle_locked_project()

        if self.config.debug is False:
            print(BColors.OKGREEN)

        if self.config.dryrun:
            xwarn("<<<< DRYRUN, do not start RMS >>>>")
            print(BColors.ENDC)
            return 0

        logger.debug(f"Execution environment: \n{self._exec_env}")
        rms_process = subprocess.run(pre_args + args, check=True)
        print(BColors.ENDC)
        return rms_process.returncode

    def run(self) -> int:
        """Main executor function"""
        self.showinfo()
        status = self._exec_rms()

        logger.debug("Status from subprocess: %s", status)

        print(
            BColors.BOLD,
            "\nRunning <runrms>. Type <runrms -h> for help\n",
            BColors.ENDC,
        )
        if not self.config.dryrun:
            self.runlogger()
        return status

    def runlogger(self) -> None:
        """Add a line to 'interactive_usage_log' defined in the site configuration.

        This log is structured as a csv with the following columns:

            date, time, user, host, full_rms_exe, commandline_options
        """
        if (
            not self.config.site_config.interactive_usage_log
            or not (usage_log := self.config.site_config.interactive_usage_log).exists()
        ):
            return

        now = datetime.datetime.now()
        nowtime = now.strftime("%Y-%m-%d,%H:%M:%S")
        user = getpass.getuser()
        host = platform.node()

        lline = "{},{},{},{},{},{}\n".format(
            nowtime, user, host, "client", self.config.executable, self.command
        )

        with open(usage_log, "a") as logg:
            logg.write(lline)

        logger.debug("Logging usage to %s:", str(usage_log))
        logger.debug(lline)

    def showinfo(self) -> None:
        """Show info on RMS project."""
        fmt = "{0:30s}: {1}"
        fmt_two = "{0:30s}: {1} {2}"
        print("=" * shutil.get_terminal_size((132, 20)).columns)
        print(f"Script runrms version {__version__}")
        print("=" * shutil.get_terminal_size((132, 20)).columns)
        print(fmt.format("Setup for runrms", self.config.site_config_file))
        print(fmt.format("Current default version", self.config.site_config.default))
        print(fmt.format("RMS version requested", self.config.version_given))
        print(fmt.format("RMS version using", self.config.version))

        if self.config.project:
            print(fmt.format("RMS project version", self.config.project.master.version))
            print(fmt.format("Project name", self.config.project.name))
            print(fmt.format("Last saved by", self.config.project.master.user))
            print(
                fmt_two.format(
                    "Last saved date & time",
                    self.config.project.master.date,
                    self.config.project.master.time,
                )
            )
            print(fmt.format("Locking info", self.config.project.lockfile))
            print(fmt.format("RMS fileversion", self.config.project.master.fileversion))
            print(fmt.format("RMS variant", self.config.project.master.variant))

        print(fmt.format("System PYTHONPATH", self.pre_env.get("PYTHONPATH", "")))
        order = "first"
        print(fmt.format(f"PYTHONPATH added as {order}", self.config.env.PYTHONPATH))
        print(fmt.format("RMS plugins path", self.config.env.RMS_PLUGINS_LIBRARY))
        print(fmt.format("TCL/TK path", self.config.env.TCL_LIBRARY))
        print(fmt.format("APS_TOOLBOX path", self.config.env.APS_TOOLBOX_PATH or ""))
        print(fmt.format("RMS DPI scaling", self.config._dpi_scaling))
        print(fmt.format("RMS executable", self.config.executable))
        print("=" * shutil.get_terminal_size((132, 20)).columns)
        print("=" * shutil.get_terminal_size((132, 20)).columns)

        expected_extension = f"rms{self.config.version}"
        if self.config.project and not self.config.project.name.endswith(
            expected_extension
        ):
            proj_name = self.config.project.name
            given_extension = proj_name.split(".", 1)[-1]

            if "rms" in proj_name:
                proj_ext_version = proj_name.split("rms")[-1]
                given_extension = f"rms{proj_ext_version}"

            xalert(
                "NOTE: Project name extension inconsistent with RMS version. "
                f"Expected <{expected_extension}>, got <{given_extension}>",
            )
